#include <windows.h>

VS_VERSION_INFO		VERSIONINFO
  FILEVERSION		1,5,30,0
  PRODUCTVERSION	1,5,30,0
  FILEFLAGSMASK		VS_FFI_FILEFLAGSMASK
#ifdef _DEBUG
  FILEFLAGS		1
#else
  FILEFLAGS		0
#endif
  FILEOS		VOS_DOS_WINDOWS32
  FILETYPE		VFT_DLL
  FILESUBTYPE		0	// not used
BEGIN
  BLOCK "StringFileInfo"
  BEGIN
    BLOCK "040904b0"
    //language ID = U.S. English, char set = Windows, Multilingual
    BEGIN
      VALUE "FileDescription",	"MATLAB MAT file I/O library\0"
      VALUE "FileVersion",	"1.5.30\0"
      VALUE "InternalName",	"libmatio.dll\0"
      VALUE "LegalCopyright",	"(C) 2005-2026, Christopher C. Hulbert and the matio contributors\0"
      VALUE "OriginalFilename",	"libmatio.dll\0"
      VALUE "ProductName",	"libmatio\0"
      VALUE "ProductVersion",	"1.5.30\0"
      VALUE "Comments",	"\0"
    END
  END
  BLOCK "VarFileInfo"
  BEGIN
    VALUE "Translation", 0x0409, 1252
  END
END
